/* sgrotum - an email SiGnature ROTUMbulator
 *
 * $Id: sgrotum.c,v 1.9 2004/07/27 05:44:00 trm Exp trm $
 * 
 * Copyright (C) 1999 - 2004, Timothy Musson, trmusson@ihug.co.nz
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA  02111-1307, USA.
 *
 * http://www.gnu.org/copyleft/gpl.html
 */


#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>
#include <string.h>
#include <glib.h>
#include <getopt.h>


#ifdef ENABLE_NLS
#  include <libintl.h>
#  define _(String) gettext(String)
#  ifdef gettext_noop
#    define N_(String) gettext_noop(String)
#  else
#    define N_(String) (String)
#  endif
#else
#  define _(String) (String)
#  define N_(String) (String)
#  define textdomain(String) (String)
#  define gettext(String) (String)
#  define dgettext(Domain,String) (String)
#  define dcgettext(Domain,String,Type) (String)
#  define bindtextdomain(Domain,Directory) (Domain)
#endif /* ENABLE_NLS */

#ifndef PACKAGE
#define PACKAGE "sgrotum"
#endif
#ifndef VERSION
#define VERSION "unknown"
#endif


typedef enum {
	ALIGN_LEFT = 0,
	ALIGN_RIGHT,
	ALIGN_CENTER
} Align;


typedef enum {
	ROTMODE_ROTATE = 0,
	ROTMODE_RANDOM
} RotMode;


typedef enum {
	PREFS_NONE = 0,
	PREFS_TARGET,
	PREFS_DASH,
	PREFS_ROTMODE,
	PREFS_QUOTES,
	PREFS_TEMPLATES,
	PREFS_CURRENT
} PrefState;


typedef struct _Sgr Sgr;
typedef struct _Box Box;


struct _Box {
	gint        x1;             /* left */
	gint        y1;             /* top */
	gint        x2;             /* right */
	gint        y2;             /* bottom */
	gint        minlen;         /* minimum length of quote to insert */
	Align       align;          /* quote alignment within box */
	gboolean    strip;          /* wrap on space and remove leading space? */
	gchar       *env;           /* value of an environment variable */
	gchar       *fname_quotes;  /* file from which to read quotes */
	gboolean    cmd;            /* treat fname_quotes as a command */
	GList       *quotes;        /* list (gchar*) of quotes that fit x1,y1,x2,y2 */
	gint        id;             /* box number (used in error messages) */
};


struct _Sgr {
	gboolean    dash;           /* prefix result with '-- \n'? */
	GList       *lines;         /* list (GString*) of lines being worked on */
	gchar       *fname_prefs;   /* name of configuration file */
	gchar       *fname_template;/* name of selected signature template file */
	gchar       *fname_quotes;  /* name of default quote file */
	gchar       *fname_dest;    /* name of signature file to write */
	GList       *template_list; /* list (gchar*) of template file names */
	GList       *pref;          /* list (gchar*) of lines in fname_prefs */
	gint        current;        /* currently selected signature template id */
	RotMode     rotmode;        /* how to update Sgr.current */
	GList       *box;           /* list (Box*) of quote boxes */
};


static gchar *attrs[] = {
	"x1",
	"y1",
	"x2",
	"y2",
	"align",
	"minlen",
	"strip",
	"file",
	"env",
	"cmd"
};


typedef enum {
	ATTR_X1 = 0,
	ATTR_Y1,
	ATTR_X2,
	ATTR_Y2,
	ATTR_ALIGN,
	ATTR_MINLEN,
	ATTR_STRIP,
	ATTR_FILE,
	ATTR_ENV,
	ATTR_CMD,
	ATTR_ERROR
} AttrType;


static struct option opts[] = {
	{ "help",      no_argument       , NULL, 'h' },
	{ "no-write",  no_argument       , NULL, 'n' },
	{ "show",      no_argument       , NULL, 's' },
	{ "template",  required_argument , NULL, 't' },
	{ "version",   no_argument       , NULL, 'v' },
	{ 0, 0, 0, 0 }
};


static gboolean do_write = TRUE;
static gboolean do_show = FALSE;

static const gchar *progname;


#ifndef __USE_GNU
#define GETLINE_GROW_SIZE 256

/**
 * getline:
 *
 * Reads a line (including '\n') from stream into *lineptr, allocating
 * space as necessary. The resulting string will be '\0' terminated.
 * Returns the number of chars read, or -1 if none.
 */
static ssize_t
getline (gchar **lineptr, size_t *size, FILE *stream)
{
	size_t n = 0;
	gint ch = 0;
	gchar *pos;

	if (lineptr == NULL || size == NULL || stream == NULL) return -1;
	if (*lineptr == NULL) {
		*size = GETLINE_GROW_SIZE;
		*lineptr = g_malloc (*size);
	}

	pos = *lineptr;
	while (ch != '\n' && (ch = fgetc (stream)) != EOF) {
		if (n == *size) {
			*size = *size + GETLINE_GROW_SIZE;
			*lineptr = g_realloc (*lineptr, *size);
			pos = *lineptr + n;
		}
		*pos = ch;
		pos++;
		n++;
	}
	*pos = '\0';

	if (n > 0) return n;
	return -1;
}
#endif /* ifndef __USE_GNU */


static FILE*
my_fopen (const gchar *fname, gchar t)
{
	FILE *f;
	gchar typestr[2];

	sprintf (typestr, "%c", t);
	f = fopen (fname, typestr);
	if (f == NULL) {
		switch (t) {
		case 'r':
			g_printerr (_("%s: unable to read file: %s\n"), PACKAGE, fname);
			break;
		case 'w':
			g_printerr (_("%s: unable to write file: %s\n"), PACKAGE, fname);
			break;
		default:
			break;
		}
	}
	return f;
}


static gchar*
expand_tilde (const gchar *fname)
{
	gchar *fname_new;
	const gchar *home;

	if (fname == NULL) return NULL;
	if (fname[0] != '~') {
		fname_new = g_strdup (fname);
	}
	else {
		home = g_getenv ("HOME");
		if (home == NULL) {
			g_printerr ("%s: getenv failed for HOME\n", PACKAGE);
			return NULL;
		}
		fname_new = g_strdup_printf ("%s%s", home, &fname[1]);
	}
	return fname_new;
}


static const gchar*
skip_space (Box *b, const gchar *str)
{
	while (isspace (*str)) str++;
	if (*str == '\0') {
		g_printerr ("%s: unexpected end of string in box %d\n", PACKAGE, b->id);
		return NULL;
	}
	return str;
}


static const gchar*
skip_n_chars (Box *b, const gchar *str, gint n)
{
	while (n > 0 && *str != '\0') {str++; n--;}
	if (n != 0) {
		g_printerr ("%s: unexpected end of string in box %d\n", PACKAGE, b->id);
		return NULL;
	}
	return str;
}


static gboolean
keep_str (const gchar *str)
{
	/* Return TRUE unless str is blank */
	while (isspace (*str)) str++;
	return *str != '#' && *str != '\0';
}


static AttrType
identify_attr (const gchar *str, gint *len)
{
	AttrType attr;

	for (attr = ATTR_X1; attr < ATTR_ERROR; attr++) {
		*len = strlen (attrs[attr]);
		if (g_ascii_strncasecmp (str, attrs[attr], *len) == 0) {
			if (str[*len] == '=' || isspace (str[*len])) {
				return attr;
			}
		}
	}
	*len = 0;
	return ATTR_ERROR;
}


static Box*
box_new (Sgr *s, gint id)
{
	Box *b = g_new (Box, 1);
	b->id     = id;
	b->x1     = 1;
	b->y1     = 1;
	b->x2     = 70;
	b->y2     = 4;
	b->minlen = 0;
	b->align  = ALIGN_LEFT;
	b->strip  = TRUE;
	b->quotes = NULL;
	b->env    = NULL;
	b->cmd    = FALSE;
	b->fname_quotes = expand_tilde (s->fname_quotes);

	return b;
}


static void
box_free (Box *b)
{
	if (b == NULL) return;
	g_free (b->fname_quotes);
	g_free (b->env);
	if (b->quotes != NULL) {
		g_list_foreach (b->quotes, (GFunc)g_free, NULL);
		g_list_free (b->quotes);
	}
	g_free (b);
}


static Sgr*
sgrotum_new (void)
{
	Sgr *s = g_new0 (Sgr, 1);
	s->dash = TRUE;
	s->rotmode = ROTMODE_RANDOM;
	s->current = 0;

	return s;
}


static void
sgrotum_free (Sgr *s)
{
	GList *l;

	if (s == NULL) return;

	if (s->box != NULL) {
		g_list_foreach (s->box, (GFunc)box_free, NULL);
		g_list_free (s->box);
	}

	l = s->lines;
	while (l != NULL) {
		g_string_free ((GString*)l->data, TRUE);
		l = l->next;
	}
	g_list_free (s->lines);

	if (s->pref != NULL) {
		g_list_foreach (s->pref, (GFunc)g_free, NULL);
		g_list_free (s->pref);
	}

	if (s->template_list != NULL) {
		g_list_foreach (s->template_list, (GFunc)g_free, NULL);
		g_list_free (s->template_list);
	}

	g_free (s->fname_prefs);
	g_free (s->fname_quotes);
	g_free (s->fname_template);
	g_free (s->fname_dest);
	g_free (s);
}


static void
cleanup (Sgr *s, gint status)
{
	sgrotum_free (s);
	exit (status == 0 ? EXIT_SUCCESS : EXIT_FAILURE);
}


static void
write_signature (Sgr *s)
{
	/* Output each line of s->lines */

	FILE *f;
	gchar *fname;
	GList *l;

	/* if there's a target file and write is TRUE, write to the target... */
	if (do_write && s->fname_dest != NULL) {

		fname = expand_tilde (s->fname_dest);

		f = my_fopen (fname, 'w');
		if (f == NULL) {
			g_free (fname);
			return;
		}

		if (s->dash) fprintf (f, "-- \n");

		l = s->lines;
		while (l != NULL) {
			fprintf (f, "%s\n", ((GString*)l->data)->str);
			l = l->next;
		}
		fclose (f);
		g_free (fname);
	}

	/* if there's no target file or show is TRUE, write to stdout... */
	if (do_show || s->fname_dest == NULL) {
		if (s->dash) fprintf (stdout, "-- \n");
		l = s->lines;
		while (l != NULL) {
			fprintf (stdout, "%s\n", ((GString*)l->data)->str);
			l = l->next;
		}
	}
}


static void
write_prefs (Sgr *s)
{
	/* Update the "current" section of the preference file */

	FILE *f;
	GList *l;

	if (s->rotmode == ROTMODE_RANDOM) return;
	if (g_list_length (s->template_list) < 2) return;

	f = my_fopen (s->fname_prefs, 'w');
	if (f == NULL) return;

	l = s->pref;
	while (l != NULL) {
		if (g_ascii_strncasecmp ("[current]", (gchar*)l->data, 9) == 0) break;
		fprintf (f, "%s", (gchar*)l->data);
		l = l->next;
	}
	fprintf (f, "[current]\n%d\n", s->current);

	fclose (f);
}


static gboolean
read_prefs (Sgr *s, const gchar *fname)
{
	FILE      *f;
	gint      len;
	guint     bufsize = 0;
	gchar     **buf, *workdir;
	PrefState state = PREFS_NONE;


	s->fname_prefs = expand_tilde (fname);

	if (!g_file_test (s->fname_prefs, G_FILE_TEST_IS_REGULAR)) {
		g_printerr (_("%s: not a regular file: %s\n"), PACKAGE, s->fname_prefs);
		return FALSE;
	}

	f = my_fopen (s->fname_prefs, 'r');
	if (f == NULL) return FALSE;

	workdir = g_path_get_dirname (s->fname_prefs);
	if (workdir != NULL) {
		chdir (workdir);
		g_free (workdir);
	}

	buf = g_malloc (sizeof (gchar*));
	*buf = NULL;

	do {
		len = getline (buf, &bufsize, f);
		if (len != -1) {
			s->pref = g_list_append (s->pref, g_strdup (*buf));
	 	}
		if (len > 1 && keep_str (*buf)) {
			if (buf[0][len - 1] == '\n') buf[0][len - 1] = '\0';
			if (g_ascii_strcasecmp (*buf, "[target]") == 0) {
				state = PREFS_TARGET;
			}
			else if (g_ascii_strcasecmp (*buf, "[dash]") == 0) {
				state = PREFS_DASH;
			}
			else if (g_ascii_strcasecmp (*buf, "[mode]") == 0) {
				state = PREFS_ROTMODE;
			}
			else if (g_ascii_strcasecmp (*buf, "[quotes]") == 0) {
				state = PREFS_QUOTES;
			}
			else if (g_ascii_strcasecmp (*buf, "[templates]") == 0) {
				state = PREFS_TEMPLATES;
			}
			else if (g_ascii_strcasecmp (*buf, "[current]") == 0) {
				state = PREFS_CURRENT;
			}
			else if (*buf[0] == '[') {
				g_printerr ("%s: unknown section in preference file ('%s')\n",
				            PACKAGE, *buf);
				state = PREFS_NONE;
			}
			else {
				switch (state) {
				case PREFS_TARGET:
					g_free (s->fname_dest);
					s->fname_dest = g_strdup (*buf);
					break;
				case PREFS_CURRENT:
					s->current = atoi (*buf);
					break;
				case PREFS_DASH:
					if (g_ascii_strcasecmp (*buf, "yes") == 0 ||
					    g_ascii_strcasecmp (*buf, "on") == 0) {
						s->dash = TRUE;
					}
					else if (g_ascii_strcasecmp (*buf, "no") == 0 ||
						g_ascii_strcasecmp (*buf, "off") == 0) {
						s->dash = FALSE;
					}
					else {
						g_printerr ("%s: bad DASH section in preference file\n", PACKAGE);
					}
					break;
				case PREFS_ROTMODE:
					if (g_ascii_strcasecmp (*buf, "rotate") == 0) {
						s->rotmode = ROTMODE_ROTATE;
					}
					else if (g_ascii_strcasecmp (*buf, "random") == 0) {
						s->rotmode = ROTMODE_RANDOM;
					}
					else {
						g_printerr ("%s: bad MODE section in preference file\n", PACKAGE);
					}
					break;
				case PREFS_QUOTES:
					g_free (s->fname_quotes);
					s->fname_quotes = g_strdup (*buf);
					break;
				case PREFS_TEMPLATES:
					s->template_list = g_list_append (s->template_list, g_strdup (*buf));
					break;
				case PREFS_NONE:
					break;
				}
			}
		}
	} while (len != -1);

	fclose (f);
	g_free (*buf);
	g_free (buf);

	if (s->template_list == NULL) {
		g_printerr (_("%s: no template files\n"), PACKAGE);
		return FALSE;
	}
	return TRUE;
}


static const gchar*
insert_text (Box *b, GString *gstr, const gchar *str, gchar *buf)
{
	/* Copy a maximum of b->x2 chars from str into gstr */

	gint x, x_space, x_max, backoff, count, offset;

	if (b->strip) while (isspace (*str)) str++;
	if (*str == '\0') return str;
	if (b->x2 > gstr->len) {
		/* pad gstr->str with trailing space chars */
		gchar *padstr = g_strnfill (b->x2 - gstr->len, ' ');
		g_string_append (gstr, padstr);
		g_free (padstr);
	}

	x_max = b->x2 - b->x1;

	/* copy chars into buf, stripping out excess space chars */
	count = 0;
	x = 0;
	while (*str != '\0' && count <= x_max) {
		if (b->strip && isspace (*str)) {
			buf[count] = ' ';
			while (isspace (*str)) str++;
			x_space = x;
			if (*str == '\0') break;
		}
		else {
			buf[count] = *str;
			str++;
		}
		count++;
		x++;
	}

	/* don't split a word */
	if (b->strip && *str != '\0' && !isspace (*str)) {
		backoff = x - x_space;
		str = str - backoff;
		count = count - backoff;
	}

	/* alignment offset */
	switch (b->align) {
	case ALIGN_LEFT:
		offset = 0;
		break;
	case ALIGN_RIGHT:
		offset = 1 + x_max - count;
		break;
	case ALIGN_CENTER:
		offset = (1 + x_max - count) / 2;
		break;
	}

	/* copy buf into gstr->str */
	memcpy (&gstr->str[b->x1 - 1] + offset, buf, count);

	/* ditch trailing space chars */
	x = gstr->len;
	while (isspace (gstr->str[x - 1])) x--;
	g_string_truncate (gstr, x);

	return str;
}


static void
insert_quote (Sgr *s, Box *b, const gchar *str)
{
	/* For each line in s->lines, insert str according to the layout
	 * given by b->x1, b->y1, b->x2, b->y2. Add extra lines to
	 * s->lines if necessary
	 */

	GList *l;
	gint i, n_lines, n_lines_needed;
	GString *gstr;
	gchar *buf;


	if (str == NULL) return;

	buf = g_malloc (b->x2 + 1);

	n_lines = g_list_length (s->lines);
	n_lines_needed = MAX (n_lines, b->y1);
	for (i = n_lines; i < n_lines_needed; i++) {
		s->lines = g_list_append (s->lines, g_string_new (""));
	}

	l = g_list_nth (s->lines, b->y1 - 1);
	while (l != NULL) {
		gstr = (GString*)l->data;
		str = insert_text (b, gstr, str, buf);
		if (b->strip) while (isspace (*str)) str++;
		if (*str != '\0' && l->next == NULL) {
			s->lines = g_list_append (s->lines, g_string_new (""));
		}
		l = l->next;
	}

	g_free (buf);
}


static const gchar*
pick_quote (Box *b)
{
	gint n;

	if (b->quotes == NULL) return NULL;
	n = g_random_int_range (0, g_list_length (b->quotes));
	return (gchar*)g_list_nth (b->quotes, n)->data;
}


static gboolean
quote_fits_box (Box *b, const gchar *str, gint len)
{
	/* Return TRUE if str is going to fit the requirements of Box b */

	gint max_chars, x_max, y_max;
	gint n_chars, len_word, x, y;


	/* too short? */
	if (len < b->minlen) return FALSE;

	/* too long? */
	x_max = 1 + b->x2 - b->x1; /* max chars per line */
	y_max = 1 + b->y2 - b->y1; /* max number of lines */
	max_chars = x_max * y_max; /* max chars in box */

	if (!b->strip) return len <= max_chars;

	while (isspace (*str)) str++;

	x = 0;
	y = 0;
	n_chars = 0;
	while (*str != '\0' && n_chars < max_chars) {

		/* get next word length */
		len_word = 0;
		while (*str != '\0' && !isspace (*str)) {
			str++;
			len_word++;
			if (len_word > x_max) return FALSE;
		}

		if (x + len_word > x_max) {
			/* wrap */
			y++;
			if (y == y_max) return FALSE;
			x = len_word;
			if (isspace (*str)) n_chars--;
		}
		else {
			x = x + len_word;
		}
		n_chars = n_chars + len_word;

		/* space */
		while (isspace (*str)) str++;
		x++;
		if (*str != '\0') n_chars++;

	}

	return *str == '\0' && n_chars >= b->minlen;
}


static gboolean
read_quotes (Sgr *s, Box *b)
{
	/* Read the quote file, adding each suitable quote to b->quotes */

	FILE    *f;
	gint    len;
	guint   bufsize = 0;
	gchar   **buf;


	if (b->fname_quotes == NULL) {
		g_printerr (_("%s: no quote file specified for template %d, box %d\n"),
		            PACKAGE, s->current, b->id);
		return FALSE;
	}

#if HAVE_POPEN
	if (b->cmd) {
		f = popen (b->fname_quotes, "r");
		if (f == NULL) return FALSE;
	}
	else {
		f = my_fopen (b->fname_quotes, 'r');
		if (f == NULL) return FALSE;
	}
#else
	f = my_fopen (b->fname_quotes, 'r');
	if (f == NULL) return FALSE;
#endif

	buf = g_malloc (sizeof (gchar*));
	*buf = NULL;

	do {
		len = getline (buf, &bufsize, f);
		if (len > 0 && keep_str (*buf)) {
			if (buf[0][len - 1] == '\n') buf[0][len - 1] = '\0';
			if (quote_fits_box (b, *buf, len - 1)) {
				b->quotes = g_list_append (b->quotes, g_strdup (*buf));
			}
		}
	} while (len != -1);

#if HAVE_POPEN
	if (b->cmd) {
		pclose (f);
	}
	else {
		fclose (f);
	}
#else
	fclose (f);
#endif

	g_free (*buf);
	g_free (buf);

	if (b->quotes == NULL) {
		g_printerr (_("%s: no quote fits template %d, box %d\n"),
		            PACKAGE, s->current, b->id);
	}

	return b->quotes != NULL;
}


static Box*
parse_layout (Sgr *s, const gchar *layout, gint id, gboolean *err)
{
	/* Return a new Box with values derived from the layout string.
	 * The layout string is passed in after the "[quote" part, so it
	 * looks something like " x1=1 y1=3 x2=70 y2=3]".
	 */

	Box *b;
	AttrType attr;
	const gchar *str = layout, *beg;
	gchar *valstr;
	gint len;


	b = box_new (s, id);

	if (*str == ']') {
		/* stick with defaults */
		*err = FALSE;
		return b;
	}

	*err = TRUE; /* till success */

	if (!isspace (*str)) {
		g_printerr ("box %d: missing space before attribute name\n", id);
		return b;
	}

	while (*str != '\0' && *str != ']') {

		/* expect and skip space */
		str = skip_space (b, str);
		if (str == NULL) return b;

		/* expect attribute name or ']' */
		if (*str == ']') {
			*err = FALSE;
			return b;
		}
		attr = identify_attr (str, &len);
		if (attr == ATTR_ERROR) {
			g_printerr ("box %d: unknown attribute name\n", id);
			return b;
		}
		str = skip_n_chars (b, str, len);
		if (str == NULL) return b;

		/* expect '=', possibly space padded */
		str = skip_space (b, str);
		if (str == NULL) return b;
		if (*str != '=') {
			g_printerr ("box %d: expected '=' after %s\n", id, attrs[attr]);
			return b;
		}
		str++;
		str = skip_space (b, str);
		if (str == NULL) return b;

		/* expect attribute value, possibly ' or " quoted */
		if (*str == '\'' || *str == '"') {
			gchar ch = *str;
			str++;
			beg = str;
			while (*str != ch && *str != '\0') str++;
			if (*str == '\0') {
				g_printerr ("box %d: unexpected end of string\n", id);
				return b;
			}
			else if (str - beg == 0) {
				g_printerr ("box %d: missing value for %s\n", id, attrs[attr]);
				return b;
			}
			valstr = g_strndup (beg, str - beg);
			str++;
		}
		else {
			beg = str;
			while (*str != ']' && *str != '\0' && !isspace (*str)) str++;
			if (*str == '\0') {
				g_printerr ("box %d: unexpected end of string\n", id);
				return b;
			}
			else if (str - beg == 0) {
				g_printerr ("box %d: missing value for %s\n", id, attrs[attr]);
				return b;
			}
			valstr = g_strndup (beg, str - beg);
		}

		/* store value according to attribute */
		switch (attr) {
		case ATTR_X1:
			b->x1 = atoi (valstr);
			if (b->x1 < 1) {
				g_printerr ("box %d: bad coordinate value\n", id);
				g_free (valstr);
				return b;
			}
			break;
		case ATTR_Y1:
			b->y1 = atoi (valstr);
			if (b->y1 < 1) {
				g_printerr ("box %d: bad coordinate value\n", id);
				g_free (valstr);
				return b;
			}
			break;
		case ATTR_X2:
			b->x2 = atoi (valstr);
			if (b->x2 < 1) {
				g_printerr ("box %d: bad coordinate value\n", id);
				g_free (valstr);
				return b;
			}
			break;
		case ATTR_Y2:
			b->y2 = atoi (valstr);
			if (b->y2 < 1) {
				g_printerr ("box %d: bad coordinate value\n", id);
				g_free (valstr);
				return b;
			}
			break;
		case ATTR_ALIGN:
			if (g_ascii_strcasecmp (valstr, "left") == 0) {
				b->align = ALIGN_LEFT;
			}
			else if (g_ascii_strcasecmp (valstr, "right") == 0) {
				b->align = ALIGN_RIGHT;
			}
			else if (g_ascii_strcasecmp (valstr, "center") == 0 ||
				     g_ascii_strcasecmp (valstr, "centre") == 0) {
				b->align = ALIGN_CENTER;
			}
			else {
				g_printerr ("box %d: bad value for align\n", id);
				g_free (valstr);
				return b;
			}
			break;
		case ATTR_MINLEN:
			b->minlen = atoi (valstr);
			if (b->minlen < 0) {
				g_printerr ("box %d: bad minlen value\n", id);
				g_free (valstr);
				return b;
			}
			break;
		case ATTR_STRIP:
			if (g_ascii_strcasecmp (valstr, "yes") == 0) {
				b->strip = TRUE;
			}
			else if (g_ascii_strcasecmp (valstr, "no") == 0) {
				b->strip = FALSE;
			}
			else {
				g_printerr ("box %d: bad strip value\n", id);
				g_free (valstr);
				return b;
			}
			break;
		case ATTR_FILE:
			g_free (b->fname_quotes);
			b->fname_quotes = expand_tilde (valstr);
			b->cmd = FALSE;
			break;
		case ATTR_CMD:
#if HAVE_POPEN
			g_free (b->fname_quotes);
			b->fname_quotes = g_strdup (valstr);
			b->cmd = TRUE;
#else
			g_printerr ("box %d: the 'cmd' option is not supported\n", id);
			g_free (valstr);
			return b;
#endif
			break;
		case ATTR_ENV:
			b->env = g_strdup (g_getenv (valstr));
			if (b->env == NULL) {
				g_printerr ("box %d: couldn't get a value for '%s'\n",
				            id, valstr);
				g_free (valstr);
				return b;
			}
			break;
		case ATTR_ERROR:
			break;
		}
		g_free (valstr);

		/* expect ']' or space */
		if (*str != ']' && !isspace (*str)) {
			g_printerr ("box %d: expected space or ']' after '%s' value\n",
			            id, attrs[attr]);
			return b;
		}
	}

	if (*str != ']') {
		g_printerr ("box %d: expected ']' at end of line\n", id);
		return b;
	}

	if (b->x2 < b->x1) {
		g_printerr ("box %d: x2 must be >= x1\n", id);
		return b;
	}
	if (b->y2 < b->y1) {
		g_printerr ("box %d: y2 must be >= y1\n", id);
		return b;
	}

	*err = FALSE;

	return b;
}


static gboolean
read_template (Sgr *s)
{
	/* Read the selected template file, storing each line in s->lines */

	FILE    *f;
	gint    len;
	guint   bufsize = 0;
	gchar   **buf;


	f = my_fopen (s->fname_template, 'r');
	if (f == NULL) return FALSE;

	buf = g_malloc (sizeof (gchar*));
	*buf = NULL;

	do {
		len = getline (buf, &bufsize, f);
		if (len > 0) {
			if (buf[0][len - 1] == '\n') buf[0][len - 1] = '\0';
			s->lines = g_list_append (s->lines, g_string_new (*buf));
		}
	} while (len != -1);

	fclose (f);
	g_free (*buf);
	g_free (buf);
	return TRUE;
}


static gboolean
pick_template (Sgr *s, gint specific)
{
	gint n_sigs;

	n_sigs = g_list_length (s->template_list);
	if (n_sigs == 0) {
		g_printerr (_("%s: no template files\n"), PACKAGE);
		return FALSE;
	}

	/* pick a new template, s->current */

	if (specific > -1) {
		if (specific >= n_sigs) {
			g_printerr (_("%s: no such template '%d'\n"), PACKAGE, specific);
			return FALSE;
		}
		s->current = specific;
	}
	else if (s->rotmode == ROTMODE_RANDOM) {
		s->current = g_random_int_range (0, n_sigs);
	}
	else {
		/* ROTMODE_ROTATE */
		s->current = s->current + 1;
		if (s->current >= n_sigs) s->current = 0;
	}

	/* get the selected template's filename */

	s->fname_template = expand_tilde
	  ((gchar*)g_list_nth (s->template_list, s->current)->data);

	return TRUE;
}


static gboolean
sgrotum (Sgr *s, gint specific)
{
	GString *gstr_layout;
	GList *l;
	Box *b;
	gint id = 1;
	gboolean err;


	if (!pick_template (s, specific)) return FALSE;
	if (!read_template (s)) return FALSE;

	if (s->lines != NULL) {
		gstr_layout = (GString*)s->lines->data;
	}
	else {
		gstr_layout = NULL;
	}

	/* add a Box for each "[quote ...]" in the selected template */

	while (gstr_layout != NULL &&
	       g_ascii_strncasecmp (gstr_layout->str, "[quote", 6) == 0) {
		b = parse_layout (s, &gstr_layout->str[6], id, &err);
		if (b == NULL || err) {
			g_printerr (_("%s: error while working on %s\n"),
			            PACKAGE, s->fname_template);
			box_free (b);
			return FALSE;
		}
		id++;

		s->box = g_list_append (s->box, b);

		/* remove this [quote ...] line from s->lines */
		s->lines = g_list_remove (s->lines, gstr_layout);
		g_string_free (gstr_layout, TRUE);
		if (s->lines != NULL) {
			gstr_layout = (GString*)s->lines->data;
		}
		else {
			gstr_layout = NULL;
		}
	}

	/* paste a quote into each box */

	l = s->box;
	while (l != NULL) {

		b = (Box*)l->data;

		if (b->env == NULL) {
			/* it's a normal quote */
			if (!read_quotes (s, b)) return FALSE;
			insert_quote (s, b, pick_quote (b));
		}
		else {
			/* it's the value of an environment variable */
			if (!quote_fits_box (b, b->env, strlen (b->env))) {
				g_printerr ("%s: env value doesn't fit template %d, box %d\n",
				            PACKAGE, s->current, b->id);
				return FALSE;
			}
			insert_quote (s, b, b->env);
		}
		l = l->next;
	}
	return TRUE;
}


static void
usage (void)
{
	g_print (_("\nUsage: %s [OPTION] [PREFERENCE_FILE]\n\n"), progname);
	g_print (_("  Generates email and newsgroup signatures\n\n"));
	g_print (_("  -h, --help           Show this usage message then exit\n" \
	           "  -n, --no-write       Don't create or modify the target signature file\n" \
	           "  -s, --show           Send the resulting signature to stdout\n" \
	           "  -t, --template NUM   Specify a signature template by number\n" \
	           "  -v, --version        Show this program's version number then exit\n\n"));
}


static void
invocation_error (void)
{
	g_printerr (_("Try `%s --help' for more information\n"), PACKAGE);
	exit (64);
}


int
main (int argc, char *argv[])
{
	const gchar *fname = NULL;
	Sgr *s;
	gint c, specific = -1;


#ifdef ENABLE_NLS
	bindtextdomain (PACKAGE, LOCALE_DIR);
	bind_textdomain_codeset (PACKAGE, "UTF-8");
	textdomain (PACKAGE);
#endif

	progname = argv[0];

	while ((c = getopt_long (argc, argv, "hnst:v", opts, NULL)) != -1) {
		switch (c) {
		case 'h':
			usage ();
			exit (EXIT_SUCCESS);
		case 'n':
			do_write = FALSE;
			break;
		case 's':
			do_show = TRUE;
			break;
		case 'v':
			g_print (VERSION "\n");
			exit (EXIT_SUCCESS);
		case 't':
			if ((sscanf (optarg, "%i", &specific)) == 0 || specific < 0) {
				g_printerr (_("%s: illegal argument '%s' (integer expected)\n"),
				            PACKAGE, optarg);
				invocation_error ();
			}
			break;
		default:
			invocation_error ();
		}
	}
	if (argc == optind) {
		fname = "~/.sgrotum/default";
	}
	if (argc > optind + 1) {
		g_printerr (_("%s: too many arguments\n"), PACKAGE);
		invocation_error ();
	}
	if (fname == NULL) {
		fname = argv[optind];
	}

	s = sgrotum_new ();

	if (!read_prefs (s, fname)) {
		cleanup (s, 1);
	}

	if (!sgrotum (s, specific)) {
		cleanup (s, 1);
	}

	write_signature (s);

	write_prefs (s);

	cleanup (s, 0);
	return 0;
}

